package rtr;

import addr.addrIP;
import addr.addrIPv4;
import cfg.cfgAll;
import java.util.ArrayList;
import java.util.List;
import pack.packHolder;
import pipe.pipeSide;
import tab.tabListing;
import tab.tabPrfxlstN;
import tab.tabRoute;
import tab.tabRouteEntry;
import tab.tabRtrmapN;
import tab.tabRtrplcN;
import util.bits;
import util.counter;
import util.debugger;
import util.logger;
import util.typLenVal;

/**
 * bgp4 speaker
 *
 * @author matecsaba
 */
public class rtrBgpSpeak implements rtrBfdClnt, Runnable {

    /**
     * connection
     */
    protected pipeSide pipe;

    private rtrBgp parent;

    private rtrBgpNeigh neigh;

    /**
     * ready to advertise
     */
    protected boolean ready2adv;

    /**
     * last received
     */
    public long lastRx;

    /**
     * uptime
     */
    public long upTime;

    /**
     * counter
     */
    public counter cntr = new counter();

    /**
     * advertised version
     */
    public int adversion = 0;

    /**
     * need full advertisement
     */
    public int needFull = 3;

    /**
     * initial full advertisement
     */
    public boolean initFull = true;

    /**
     * learned unicast prefixes
     */
    public tabRoute<addrIP> lrnUni = new tabRoute<addrIP>("rx");

    /**
     * learned multicast prefixes
     */
    public tabRoute<addrIP> lrnMlt = new tabRoute<addrIP>("rx");

    /**
     * learned flowspec prefixes
     */
    public tabRoute<addrIP> lrnFlw = new tabRoute<addrIP>("rx");

    /**
     * learned vpnuni prefixes
     */
    public tabRoute<addrIP> lrnVpnU = new tabRoute<addrIP>("rx");

    /**
     * learned vpnmulti prefixes
     */
    public tabRoute<addrIP> lrnVpnM = new tabRoute<addrIP>("rx");

    /**
     * learned vpnflow prefixes
     */
    public tabRoute<addrIP> lrnVpnF = new tabRoute<addrIP>("rx");

    /**
     * learned vpls prefixes
     */
    public tabRoute<addrIP> lrnVpls = new tabRoute<addrIP>("rx");

    /**
     * learned evpn prefixes
     */
    public tabRoute<addrIP> lrnEvpn = new tabRoute<addrIP>("rx");

    /**
     * learned mdt prefixes
     */
    public tabRoute<addrIP> lrnMdt = new tabRoute<addrIP>("rx");

    /**
     * advertised unicast prefixes
     */
    public tabRoute<addrIP> advUni = new tabRoute<addrIP>("tx");

    /**
     * advertised multicast prefixes
     */
    public tabRoute<addrIP> advMlt = new tabRoute<addrIP>("tx");

    /**
     * advertised flowspec prefixes
     */
    public tabRoute<addrIP> advFlw = new tabRoute<addrIP>("tx");

    /**
     * advertised vpnuni prefixes
     */
    public tabRoute<addrIP> advVpnU = new tabRoute<addrIP>("tx");

    /**
     * advertised vpnmulti prefixes
     */
    public tabRoute<addrIP> advVpnM = new tabRoute<addrIP>("tx");

    /**
     * advertised vpnflow prefixes
     */
    public tabRoute<addrIP> advVpnF = new tabRoute<addrIP>("tx");

    /**
     * advertised vpls prefixes
     */
    public tabRoute<addrIP> advVpls = new tabRoute<addrIP>("tx");

    /**
     * advertised evpn prefixes
     */
    public tabRoute<addrIP> advEvpn = new tabRoute<addrIP>("tx");

    /**
     * advertised mdt prefixes
     */
    public tabRoute<addrIP> advMdt = new tabRoute<addrIP>("tx");

    /**
     * currently updating unicast prefixes
     */
    private tabRoute<addrIP> currUni = null;

    /**
     * currently updating multicast prefixes
     */
    private tabRoute<addrIP> currMlt = null;

    /**
     * currently updating flowspec prefixes
     */
    private tabRoute<addrIP> currFlw = null;

    /**
     * currently updating vpn uni prefixes
     */
    private tabRoute<addrIP> currVpnU = null;

    /**
     * currently updating vpn multi prefixes
     */
    private tabRoute<addrIP> currVpnM = null;

    /**
     * currently updating vpn flow prefixes
     */
    private tabRoute<addrIP> currVpnF = null;

    /**
     * currently updating vpls prefixes
     */
    private tabRoute<addrIP> currVpls = null;

    /**
     * currently updating evpn prefixes
     */
    private tabRoute<addrIP> currEvpn = null;

    /**
     * currently updating mdt prefixes
     */
    private tabRoute<addrIP> currMdt = null;

    /**
     * currently changed prefixes
     */
    private int currChg;

    /**
     * header size
     */
    public static final int size = 19;

    /**
     * peer hold time
     */
    public int peerHold;

    /**
     * negotiated keepalive time
     */
    public int peerKeep;

    /**
     * peer router id
     */
    public addrIPv4 peerRouterID = new addrIPv4();

    /**
     * peer 32bit asn capability
     */
    public boolean peer32bitAS;

    /**
     * peer address families
     */
    public int peerAfis;

    /**
     * peer route refresh capability
     */
    public boolean peerRefresh;

    /**
     * peer graceful restart capability
     */
    public int peerGrace;

    /**
     * peer sends additional paths
     */
    public int addpathRx;

    /**
     * peer receives additional paths
     */
    public int addpathTx;

    private packHolder pckRx = new packHolder(true, true);

    private packHolder pckRh = new packHolder(true, true);

    private packHolder pckTx = new packHolder(true, true);

    private packHolder pckTh = new packHolder(true, true);

    /**
     * create bgp speaker
     *
     * @param protocol process
     * @param neighbor neighbor
     * @param socket tcp socket to use, null if none
     */
    public rtrBgpSpeak(rtrBgp protocol, rtrBgpNeigh neighbor, pipeSide socket) {
        ready2adv = false;
        parent = protocol;
        neigh = neighbor;
        pipe = socket;
        if (pipe == null) {
            return;
        }
        upTime = bits.getTime();
        lastRx = upTime;
        if (parent == null) {
            return;
        }
        new Thread(this).start();
    }

    private boolean afiMsk(int val, int safi) {
        if (safi == parent.afiUni) {
            return (val & rtrBgpParam.mskUni) != 0;
        }
        if (safi == parent.afiLab) {
            return (val & rtrBgpParam.mskLab) != 0;
        }
        if (safi == parent.afiMlt) {
            return (val & rtrBgpParam.mskMlt) != 0;
        }
        if (safi == parent.afiFlw) {
            return (val & rtrBgpParam.mskFlw) != 0;
        }
        if (safi == parent.afiVpnU) {
            return (val & rtrBgpParam.mskVpnU) != 0;
        }
        if (safi == parent.afiVpnM) {
            return (val & rtrBgpParam.mskVpnM) != 0;
        }
        if (safi == parent.afiVpnF) {
            return (val & rtrBgpParam.mskVpnF) != 0;
        }
        if (safi == parent.afiVpls) {
            return (val & rtrBgpParam.mskVpls) != 0;
        }
        if (safi == parent.afiEvpn) {
            return (val & rtrBgpParam.mskEvpn) != 0;
        }
        if (safi == parent.afiMdt) {
            return (val & rtrBgpParam.mskMdt) != 0;
        }
        return false;
    }

    /**
     * get learned
     *
     * @param safi safi to query
     * @return table
     */
    public tabRoute<addrIP> getLearned(int safi) {
        if (safi == parent.afiUni) {
            return lrnUni;
        }
        if (safi == parent.afiLab) {
            return lrnUni;
        }
        if (safi == parent.afiMlt) {
            return lrnMlt;
        }
        if (safi == parent.afiFlw) {
            return lrnFlw;
        }
        if (safi == parent.afiVpnU) {
            return lrnVpnU;
        }
        if (safi == parent.afiVpnM) {
            return lrnVpnM;
        }
        if (safi == parent.afiVpnF) {
            return lrnVpnF;
        }
        if (safi == parent.afiVpls) {
            return lrnVpls;
        }
        if (safi == parent.afiEvpn) {
            return lrnEvpn;
        }
        if (safi == parent.afiMdt) {
            return lrnMdt;
        }
        return null;
    }

    /**
     * get adverted
     *
     * @param safi safi to query
     * @return table
     */
    public tabRoute<addrIP> getAdverted(int safi) {
        if (safi == parent.afiUni) {
            return advUni;
        }
        if (safi == parent.afiLab) {
            return advUni;
        }
        if (safi == parent.afiMlt) {
            return advMlt;
        }
        if (safi == parent.afiFlw) {
            return advFlw;
        }
        if (safi == parent.afiVpnU) {
            return advVpnU;
        }
        if (safi == parent.afiVpnM) {
            return advVpnM;
        }
        if (safi == parent.afiVpnF) {
            return advVpnF;
        }
        if (safi == parent.afiVpls) {
            return advVpls;
        }
        if (safi == parent.afiEvpn) {
            return advEvpn;
        }
        if (safi == parent.afiMdt) {
            return advMdt;
        }
        return null;
    }

    /**
     * addpath rx test
     *
     * @param safi safi to test
     * @return true if addpath in use
     */
    public boolean addPthRx(int safi) {
        return afiMsk(addpathRx, safi);
    }

    /**
     * addpath tx test
     *
     * @param safi safi to test
     * @return true if addpath in use
     */
    public boolean addPthTx(int safi) {
        return afiMsk(addpathTx, safi);
    }

    /**
     * close this session
     */
    public void closeNow() {
        if (pipe != null) {
            pipe.setClose();
        }
        if (neigh.localIfc != null) {
            neigh.localIfc.bfdDel(neigh.peerAddr, this);
        }
        lrnUni = new tabRoute<addrIP>("rx");
        lrnMlt = new tabRoute<addrIP>("rx");
        lrnFlw = new tabRoute<addrIP>("rx");
        lrnVpnU = new tabRoute<addrIP>("rx");
        lrnVpnM = new tabRoute<addrIP>("rx");
        lrnVpnF = new tabRoute<addrIP>("rx");
        lrnVpls = new tabRoute<addrIP>("rx");
        lrnEvpn = new tabRoute<addrIP>("rx");
        lrnMdt = new tabRoute<addrIP>("rx");
        advUni = new tabRoute<addrIP>("tx");
        advMlt = new tabRoute<addrIP>("tx");
        advFlw = new tabRoute<addrIP>("tx");
        advVpnU = new tabRoute<addrIP>("tx");
        advVpnM = new tabRoute<addrIP>("tx");
        advVpnF = new tabRoute<addrIP>("tx");
        advVpls = new tabRoute<addrIP>("tx");
        advEvpn = new tabRoute<addrIP>("tx");
        advMdt = new tabRoute<addrIP>("tx");
        neigh.accUni = new tabRoute<addrIP>("rx");
        neigh.accMlt = new tabRoute<addrIP>("rx");
        neigh.accFlw = new tabRoute<addrIP>("rx");
        neigh.accVpnU = new tabRoute<addrIP>("rx");
        neigh.accVpnM = new tabRoute<addrIP>("rx");
        neigh.accVpnF = new tabRoute<addrIP>("rx");
        neigh.accVpls = new tabRoute<addrIP>("rx");
        neigh.accEvpn = new tabRoute<addrIP>("rx");
        neigh.accMdt = new tabRoute<addrIP>("rx");
        neigh.wilUni = new tabRoute<addrIP>("rx");
        neigh.wilMlt = new tabRoute<addrIP>("rx");
        neigh.wilFlw = new tabRoute<addrIP>("rx");
        neigh.wilVpnU = new tabRoute<addrIP>("rx");
        neigh.wilVpnM = new tabRoute<addrIP>("rx");
        neigh.wilVpnF = new tabRoute<addrIP>("rx");
        neigh.wilVpls = new tabRoute<addrIP>("rx");
        neigh.wilEvpn = new tabRoute<addrIP>("rx");
        neigh.wilMdt = new tabRoute<addrIP>("rx");
        neigh.chgUni = new tabRoute<addrIP>("chg");
        neigh.chgMlt = new tabRoute<addrIP>("chg");
        neigh.chgFlw = new tabRoute<addrIP>("chg");
        neigh.chgVpnU = new tabRoute<addrIP>("chg");
        neigh.chgVpnM = new tabRoute<addrIP>("chg");
        neigh.chgVpnF = new tabRoute<addrIP>("chg");
        neigh.chgVpls = new tabRoute<addrIP>("chg");
        neigh.chgEvpn = new tabRoute<addrIP>("chg");
        neigh.chgMdt = new tabRoute<addrIP>("chg");
        if (!ready2adv) {
            return;
        }
        ready2adv = false;
        adversion = 0;
        needFull = 3;
        initFull = true;
        parent.needFull = true;
        parent.compute.wakeup();
    }

    public void bfdPeerDown() {
        closeNow();
    }

    /**
     * check tx buffer size
     *
     * @return less than 0 means error, otherwise bytes free
     */
    public int txFree() {
        if (pipe == null) {
            return -1;
        }
        if (pipe.isClosed() != 0) {
            return -1;
        }
        if (!ready2adv) {
            return 0;
        }
        return pipe.ready2tx();
    }

    public void run() {
        try {
            doWork();
        } catch (Exception e) {
            logger.traceback(e);
        }
        closeNow();
    }

    private void doWork() {
        if (debugger.rtrBgpEvnt) {
            logger.debug("starting neighbor " + neigh.peerAddr);
        }
        pipe.timeout = neigh.holdTimer;
        peerKeep = neigh.keepAlive;
        pipe.setReady();
        pipe.wait4ready(neigh.keepAlive);
        if (pipe.isReady() != 3) {
            closeNow();
            return;
        }
        sendOpen();
        sendKeepAlive();
        int typ = packRecv(pckRx);
        if (typ == rtrBgpUtil.msgNotify) {
            logger.info("got notify from " + neigh.peerAddr);
            closeNow();
            return;
        }
        if (typ != rtrBgpUtil.msgOpen) {
            sendNotify(1, 3);
            return;
        }
        if (parseOpen(pckRx)) {
            closeNow();
            return;
        }
        logger.warn("neighbor " + neigh.peerAddr + " up");
        if (neigh.bfdTrigger) {
            neigh.localIfc.bfdAdd(neigh.peerAddr, this, "bgp");
        }
        lrnUni = new tabRoute<addrIP>("rx");
        lrnMlt = new tabRoute<addrIP>("rx");
        lrnFlw = new tabRoute<addrIP>("rx");
        lrnVpnU = new tabRoute<addrIP>("rx");
        lrnVpnM = new tabRoute<addrIP>("rx");
        lrnVpnF = new tabRoute<addrIP>("rx");
        lrnVpls = new tabRoute<addrIP>("rx");
        lrnEvpn = new tabRoute<addrIP>("rx");
        lrnMdt = new tabRoute<addrIP>("rx");
        neigh.accUni = new tabRoute<addrIP>("rx");
        neigh.accMlt = new tabRoute<addrIP>("rx");
        neigh.accFlw = new tabRoute<addrIP>("rx");
        neigh.accVpnU = new tabRoute<addrIP>("rx");
        neigh.accVpnM = new tabRoute<addrIP>("rx");
        neigh.accVpnF = new tabRoute<addrIP>("rx");
        neigh.accVpls = new tabRoute<addrIP>("rx");
        neigh.accEvpn = new tabRoute<addrIP>("rx");
        neigh.accMdt = new tabRoute<addrIP>("rx");
        ready2adv = true;
        parent.needFull = true;
        parent.compute.wakeup();
        for (;;) {
            typ = packRecv(pckRx);
            if (typ < 0) {
                sendNotify(1, 1);
                break;
            }
            lastRx = bits.getTime();
            if (typ == rtrBgpUtil.msgKeepLiv) {
                continue;
            }
            if (typ == rtrBgpUtil.msgNotify) {
                logger.info("got notify (" + pckRx.getByte(0) + "/" + pckRx.getByte(1) + ") from " + neigh.peerAddr);
                break;
            }
            if (typ == rtrBgpUtil.msgRefrsh) {
                int safi = pckRx.msbGetD(0);
                gotRefresh(safi);
                continue;
            }
            if (typ != rtrBgpUtil.msgUpdate) {
                logger.info("got unknown type (" + typ + ") from " + neigh.peerAddr);
                sendNotify(1, 3);
                break;
            }
            if (parseUpdate(pckRx, pckRh)) {
                logger.info("got malformed update from " + neigh.peerAddr);
                sendNotify(3, 1);
                break;
            }
        }
        closeNow();
        logger.error("neighbor " + neigh.peerAddr + " down");
    }

    /**
     * send one packet
     *
     * @param pck packet to send
     * @param typ type of packet
     */
    public void packSend(packHolder pck, int typ) {
        if (pipe == null) {
            return;
        }
        if (debugger.rtrBgpEvnt) {
            logger.debug("sending " + rtrBgpUtil.type2string(typ) + " to " + neigh.peerAddr);
        }
        if (neigh.monitor != null) {
            neigh.monitor.gotMessage(true, typ, this, neigh, pck.getCopy());
        }
        if (neigh.dump != null) {
            neigh.dump.gotMessage(true, typ, neigh, pck.getCopy());
        }
        pck.merge2beg();
        cntr.tx(pck);
        for (int i = 0; i < 16; i++) {
            pck.putByte(i, 0xff);
        }
        pck.msbPutW(16, pck.dataSize() + size);
        pck.putByte(18, typ);
        pck.putSkip(size);
        pck.merge2beg();
        pck.pipeSend(pipe, 0, pck.dataSize(), 2);
    }

    /**
     * receive one packet
     *
     * @param pck packet to store to
     * @return type of message, negative on error
     */
    private int packRecv(packHolder pck) {
        pck.clear();
        if (pipe == null) {
            return -1;
        }
        if (pck.pipeRecv(pipe, 0, size, 144) != size) {
            return -1;
        }
        for (int i = 0; i < 16; i++) {
            if (pck.getByte(i) != 0xff) {
                return -1;
            }
        }
        int len = pck.msbGetW(16) - size;
        int typ = pck.getByte(18);
        if (len < 0) {
            return -1;
        }
        if (len > packHolder.maxData) {
            return -1;
        }
        pck.clear();
        if (len > 0) {
            if (pck.pipeRecv(pipe, 0, len, 144) != len) {
                return -1;
            }
        }
        cntr.rx(pck);
        if (debugger.rtrBgpEvnt) {
            logger.debug("got " + rtrBgpUtil.type2string(typ) + " from " + neigh.peerAddr);
        }
        if (neigh.monitor != null) {
            neigh.monitor.gotMessage(false, typ, this, neigh, pckRx.getCopy());
        }
        if (neigh.dump != null) {
            neigh.dump.gotMessage(false, typ, neigh, pckRx.getCopy());
        }
        return typ;
    }

    /**
     * send keep alive message
     */
    public void sendKeepAlive() {
        packHolder pck = new packHolder(true, true);
        packSend(pck, rtrBgpUtil.msgKeepLiv);
    }

    /**
     * send open message
     */
    public void sendOpen() {
        List<Integer> safis = parent.mask2list(neigh.addrFams);
        packHolder pck = new packHolder(true, true);
        for (int i = 0; i < safis.size(); i++) {
            byte[] buf = new byte[4];
            bits.msbPutD(buf, 0, safis.get(i));
            rtrBgpUtil.placeCapability(pck, rtrBgpUtil.capaMultiProto, buf);
        }
        byte[] buf = new byte[4];
        bits.msbPutD(buf, 0, neigh.localAs);
        rtrBgpUtil.placeCapability(pck, rtrBgpUtil.capa32bitAsNum, buf);
        buf = new byte[0];
        rtrBgpUtil.placeCapability(pck, rtrBgpUtil.capaRouteRefresh, buf);
        safis = parent.mask2list((neigh.addpathRmode | neigh.addpathTmode) & neigh.addrFams);
        if (safis.size() > 0) {
            buf = new byte[safis.size() * 4];
            for (int i = 0; i < safis.size(); i++) {
                int o = safis.get(i);
                bits.msbPutD(buf, i * 4, rtrBgpUtil.safi2triplet(o));
                o = parent.safi2mask(o);
                int m = 0;
                if ((neigh.addpathRmode & o) != 0) {
                    m |= 1;
                }
                if ((neigh.addpathTmode & o) != 0) {
                    m |= 2;
                }
                buf[(i * 4) + 3] = (byte) m;
            }
            rtrBgpUtil.placeCapability(pck, rtrBgpUtil.capaAdditionPath, buf);
        }
        safis = parent.mask2list(neigh.graceRestart & neigh.addrFams);
        if (safis.size() > 0) {
            buf = new byte[2 + (safis.size() * 4)];
            bits.msbPutW(buf, 0, ((parent.restartTime / 1000) & 0xfff) | 0x8000);
            for (int i = 0; i < safis.size(); i++) {
                bits.msbPutD(buf, (i * 4) + 2, rtrBgpUtil.safi2triplet(safis.get(i)));
            }
            rtrBgpUtil.placeCapability(pck, rtrBgpUtil.capaGraceRestart, buf);
        }
        if (neigh.hostname) {
            buf = cfgAll.hostName.getBytes();
            byte[] tmp = new byte[1];
            tmp[0] = (byte) buf.length;
            buf = bits.byteConcat(tmp, buf);
            tmp = new byte[1];
            buf = bits.byteConcat(buf, tmp);
            rtrBgpUtil.placeCapability(pck, rtrBgpUtil.capaHostname, buf);
        }
        pck.merge2beg();
        pck.putByte(0, rtrBgpUtil.version);
        pck.msbPutW(1, rtrBgpUtil.asNum16bit(neigh.localAs));
        pck.msbPutW(3, neigh.holdTimer / 1000);
        buf = parent.routerID.getBytes();
        pck.putCopy(buf, 0, 5, buf.length);
        pck.putByte(9, pck.dataSize());
        pck.putSkip(10);
        pck.merge2beg();
        packSend(pck, rtrBgpUtil.msgOpen);
    }

    /**
     * parse open packet
     *
     * @param pck packet to use
     * @return false on success, true on error
     */
    public boolean parseOpen(packHolder pck) {
        if (pck.getByte(0) != rtrBgpUtil.version) {
            logger.info("neighbor " + neigh.peerAddr + " has bad version");
            sendNotify(2, 1);
            return true;
        }
        int i = pck.msbGetW(1);
        if ((neigh.remoteAs != -1) && (i != rtrBgpUtil.asNum16bit(neigh.remoteAs))) {
            logger.info("neighbor " + neigh.peerAddr + " in wrong (" + i + ") as");
            sendNotify(2, 2);
            return true;
        }
        peerHold = pck.msbGetW(3) * 1000;
        if (peerHold < neigh.holdTimer) {
            peerKeep = peerHold / 3;
            pipe.timeout = peerHold;
        }
        pck.getAddr(peerRouterID, 5);
        i = pck.getByte(9);
        pck.getSkip(10);
        if (i < pck.dataSize()) {
            pck.setDataSize(i);
        }
        boolean mpGot = false;
        for (;;) {
            if (pck.dataSize() < 1) {
                break;
            }
            typLenVal tlv = new typLenVal(0, 8, 8, 8, 1, 0, 2, 1, 0, 512, true);
            if (tlv.getBytes(pck)) {
                break;
            }
            if (tlv.valTyp != 2) {
                continue;
            }
            packHolder pck2 = new packHolder(true, true);
            pck2.putCopy(tlv.valDat, 0, 0, tlv.valSiz);
            pck2.putSkip(tlv.valSiz);
            pck2.merge2beg();
            for (;;) {
                tlv = new typLenVal(0, 8, 8, 8, 1, 0, 2, 1, 0, 512, true);
                if (tlv.getBytes(pck2)) {
                    break;
                }
                switch (tlv.valTyp) {
                    case rtrBgpUtil.capa32bitAsNum:
                        i = bits.msbGetD(tlv.valDat, 0);
                        if ((neigh.remoteAs != -1) && (i != neigh.remoteAs)) {
                            logger.info("neighbor " + neigh.peerAddr + " in wrong (" + bits.num2str(i) + ") as");
                            sendNotify(2, 2);
                            return true;
                        }
                        peer32bitAS = true;
                        break;
                    case rtrBgpUtil.capaMultiProto:
                        mpGot = true;
                        for (i = 0; i < tlv.valSiz; i += 4) {
                            int o = bits.msbGetD(tlv.valDat, i);
                            o = parent.safi2mask(o);
                            if (o < 1) {
                                continue;
                            }
                            peerAfis |= o;
                        }
                        break;
                    case rtrBgpUtil.capaGraceRestart:
                        for (i = 2; i < tlv.valSiz; i += 4) {
                            int o = bits.msbGetD(tlv.valDat, i);
                            o = rtrBgpUtil.triplet2safi(o);
                            o = parent.safi2mask(o);
                            if (o < 1) {
                                continue;
                            }
                            peerGrace |= o;
                        }
                        break;
                    case rtrBgpUtil.capaAdditionPath:
                        for (i = 0; i < tlv.valSiz; i += 4) {
                            int o = bits.msbGetD(tlv.valDat, i);
                            o = rtrBgpUtil.triplet2safi(o);
                            o = parent.safi2mask(o);
                            if (o < 1) {
                                continue;
                            }
                            int m = tlv.valDat[i + 3];
                            if ((m & 2) != 0) {
                                addpathRx |= o;
                            }
                            if ((m & 1) != 0) {
                                addpathTx |= o;
                            }
                        }
                        break;
                    case rtrBgpUtil.capaRouteRefresh:
                        peerRefresh = true;
                        break;
                }
            }
        }
        if (!mpGot) {
            int o = parent.safi2mask(rtrBgpUtil.safiIp4uni);
            if (o > 0) {
                peerAfis |= o;
            }
        }
        peerAfis &= neigh.addrFams;
        if (peerAfis == 0) {
            logger.info("neighbor " + neigh.peerAddr + " in wrong safi");
            sendNotify(6, 3);
            return true;
        }
        addpathRx &= neigh.addpathRmode;
        addpathTx &= neigh.addpathTmode;
        addpathRx &= neigh.addrFams;
        addpathTx &= neigh.addrFams;
        if (debugger.rtrBgpEvnt) {
            logger.debug("peer " + neigh.peerAddr + " id=" + peerRouterID + " hold=" + peerHold + " 32bitAS=" + peer32bitAS + " refresh=" + peerRefresh);
        }
        return false;
    }

    /**
     * send keep alive message
     *
     * @param err error code
     * @param sub sub code
     */
    public void sendNotify(int err, int sub) {
        packHolder pck = new packHolder(true, true);
        pck.putByte(0, err);
        pck.putByte(1, sub);
        pck.putSkip(2);
        packSend(pck, rtrBgpUtil.msgNotify);
        closeNow();
        logger.info("sent notify (" + err + "/" + sub + ") to " + neigh.peerAddr);
    }

    /**
     * got route refresh request
     *
     * @param safi safi to refresh
     */
    public void gotRefresh(int safi) {
        if (!afiMsk(peerAfis, safi)) {
            return;
        }
        tabRoute<addrIP> adverted = getAdverted(safi);
        if (adverted == null) {
            return;
        }
        adverted.clear();
        needFull = 3;
        initFull = true;
        adversion--;
        if (debugger.rtrBgpTraf) {
            logger.debug("got refresh from peer " + neigh.peerAddr + " in " + rtrBgpUtil.safi2string(safi));
        }
    }

    /**
     * send route refresh request
     *
     * @param safi safi to refresh
     */
    public void sendRefresh(int safi) {
        if (!peerRefresh) {
            return;
        }
        if (!afiMsk(peerAfis, safi)) {
            return;
        }
        tabRoute<addrIP> learned = getLearned(safi);
        if (learned == null) {
            return;
        }
        learned.clear();
        parent.needFull = true;
        parent.compute.wakeup();
        packHolder pck = new packHolder(true, true);
        pck.msbPutD(0, safi);
        pck.putSkip(4);
        packSend(pck, rtrBgpUtil.msgRefrsh);
        if (debugger.rtrBgpTraf) {
            logger.debug("sent refresh to peer " + neigh.peerAddr + " in " + rtrBgpUtil.safi2string(safi));
        }
    }

    /**
     * send update packet
     *
     * @param safi safi to update
     * @param ntry prefix to advertise
     * @param reach true to reachable, false to withdraw
     */
    public void sendUpdate(int safi, tabRouteEntry<addrIP> ntry, boolean reach) {
        if (debugger.rtrBgpTraf) {
            logger.debug("update to peer " + neigh.peerAddr + " in " + rtrBgpUtil.safi2string(safi) + ": " + (reach ? "reachable" : "withdraw") + " " + tabRtrmapN.rd2string(ntry.rouDst) + " " + ntry.prefix);
        }
        pckTx.clear();
        boolean addpath = addPthTx(safi);
        if (!reach) {
            rtrBgpUtil.createWithdraw(pckTx, pckTh, safi, addpath, ntry);
        } else {
            rtrBgpUtil.createReachable(pckTx, pckTh, safi, addpath, peer32bitAS, ntry);
        }
        packSend(pckTx, rtrBgpUtil.msgUpdate);
    }

    /**
     * parse update packet
     *
     * @param pck packet to parse
     * @param hlp helper packet
     * @return false on success, true on error
     */
    public boolean parseUpdate(packHolder pck, packHolder hlp) {
        if (debugger.rtrBgpTraf) {
            logger.debug("update from peer " + neigh.peerAddr);
        }
        currUni = new tabRoute<addrIP>("curr");
        currMlt = new tabRoute<addrIP>("curr");
        currFlw = new tabRoute<addrIP>("curr");
        currVpnU = new tabRoute<addrIP>("curr");
        currVpnM = new tabRoute<addrIP>("curr");
        currVpnF = new tabRoute<addrIP>("curr");
        currVpls = new tabRoute<addrIP>("curr");
        currEvpn = new tabRoute<addrIP>("curr");
        currMdt = new tabRoute<addrIP>("curr");
        currChg = 0;
        int prt = pck.msbGetW(0);
        pck.getSkip(2);
        prt = pck.dataSize() - prt;
        tabRouteEntry<addrIP> ntry = new tabRouteEntry<addrIP>();
        ntry.rouTyp = neigh.lower.rouTyp;
        ntry.protoNum = neigh.lower.rtrNum;
        ntry.distance = neigh.distance;
        ntry.rouSrc = neigh.peerType;
        ntry.srcRtr = neigh.peerAddr.copyBytes();
        ntry.locPref = 100;
        ntry.nextHop = neigh.peerAddr.copyBytes();
        for (;;) {
            if (pck.dataSize() <= prt) {
                break;
            }
            if (addPthRx(rtrBgpUtil.safiIp4uni)) {
                pck.getSkip(4);
            }
            tabRouteEntry<addrIP> res = rtrBgpUtil.readPrefix(rtrBgpUtil.safiIp4uni, false, pck);
            ntry.prefix = res.prefix;
            ntry.labelRem = res.labelRem;
            prefixWithdraw(rtrBgpUtil.safiIp4uni, ntry);
        }
        pck.setBytesLeft(prt);
        prt = pck.msbGetW(0);
        pck.getSkip(2);
        prt = pck.dataSize() - prt;
        for (;;) {
            if (pck.dataSize() <= prt) {
                break;
            }
            rtrBgpUtil.parseAttrib(pck, hlp);
            rtrBgpUtil.interpretAttribute(this, ntry, hlp);
        }
        pck.setBytesLeft(prt);
        for (;;) {
            if (pck.dataSize() < 1) {
                break;
            }
            if (addPthRx(rtrBgpUtil.safiIp4uni)) {
                pck.getSkip(4);
            }
            tabRouteEntry<addrIP> res = rtrBgpUtil.readPrefix(rtrBgpUtil.safiIp4uni, false, pck);
            ntry.prefix = res.prefix;
            ntry.labelRem = res.labelRem;
            prefixReach(rtrBgpUtil.safiIp4uni, ntry);
        }
        addAttribed(currUni, parent.afiUni, ntry, neigh.roumapIn, neigh.roupolIn, neigh.prflstIn);
        addAttribed(currUni, parent.afiLab, ntry, neigh.roumapIn, neigh.roupolIn, neigh.prflstIn);
        addAttribed(currMlt, parent.afiMlt, ntry, neigh.roumapIn, neigh.roupolIn, neigh.prflstIn);
        addAttribed(currFlw, parent.afiFlw, ntry, neigh.voumapIn, neigh.voupolIn, null);
        addAttribed(currVpnU, parent.afiVpnU, ntry, neigh.voumapIn, neigh.voupolIn, null);
        addAttribed(currVpnM, parent.afiVpnM, ntry, neigh.voumapIn, neigh.voupolIn, null);
        addAttribed(currVpnF, parent.afiVpnF, ntry, neigh.voumapIn, neigh.voupolIn, null);
        addAttribed(currVpls, parent.afiVpls, ntry, neigh.voumapIn, neigh.voupolIn, null);
        addAttribed(currEvpn, parent.afiEvpn, ntry, neigh.voumapIn, neigh.voupolIn, null);
        addAttribed(currMdt, parent.afiMdt, ntry, neigh.voumapIn, neigh.voupolIn, null);
        currUni = null;
        currMlt = null;
        currFlw = null;
        currVpnU = null;
        currVpnM = null;
        currVpnF = null;
        currVpls = null;
        currEvpn = null;
        currMdt = null;
        if (currChg > 0) {
            parent.compute.wakeup();
        }
        if (neigh.maxPrefixCnt < 1) {
            return false;
        }
        int i = lrnUni.size() + lrnMlt.size();
        if (i > ((neigh.maxPrefixCnt * neigh.maxPrefixPrc) / 100)) {
            logger.info("neighbor " + neigh.peerAddr + " sent " + i + " prefixes");
        }
        if (i > neigh.maxPrefixCnt) {
            sendNotify(6, 1);
        }
        return false;
    }

    private void addAttribed(tabRoute<addrIP> currAdd, int safi, tabRouteEntry<addrIP> attr, tabListing<tabRtrmapN, addrIP> roumap, tabListing<tabRtrplcN, addrIP> roupol, tabListing<tabPrfxlstN, addrIP> prflst) {
        if (!afiMsk(peerAfis, safi)) {
            return;
        }
        tabRoute<addrIP> learned = getLearned(safi);
        if (learned == null) {
            return;
        }
        tabRoute<addrIP> changed = parent.getChanged(safi);
        if (changed == null) {
            parent.needFull = true;
            currChg++;
        }
        tabRoute<addrIP> acc = new tabRoute<addrIP>("curr");
        for (int i = 0; i < currAdd.size(); i++) {
            tabRouteEntry<addrIP> cur = currAdd.get(i);
            attr.prefix = cur.prefix;
            attr.nextHop = cur.nextHop;
            attr.rouDst = cur.rouDst;
            attr.labelRem = cur.labelRem;
            attr.evpnLab = cur.evpnLab;
            cur = attr.copyBytes();
            acc.add(3, cur, false, true);
            if (learned.del(cur)) {
                continue;
            }
            currChg++;
            if (changed == null) {
                continue;
            }
            changed.add(3, cur, false, false);
        }
        if (!neigh.softReconfig) {
            tabRoute<addrIP> ok = new tabRoute<addrIP>("curr");
            tabRoute.addUpdatedTable(2, ok, acc, roumap, roupol, prflst);
            acc = ok;
        }
        for (int i = 0; i < acc.size(); i++) {
            tabRouteEntry<addrIP> ntry = acc.get(i);
            if ((safi == parent.afiUni) && (parent.flaps != null)) {
                parent.prefixFlapped(ntry.prefix, ntry.asPathStr());
            }
            if (prefixReachable(learned, ntry)) {
                continue;
            }
            currChg++;
            if (changed == null) {
                continue;
            }
            changed.add(3, ntry, false, false);
        }
    }

    /**
     * prefix withdrawal received
     *
     * @param safi safi
     * @param ntry prefix
     */
    protected void prefixWithdraw(int safi, tabRouteEntry<addrIP> ntry) {
        if (debugger.rtrBgpTraf) {
            logger.debug("withdraw " + rtrBgpUtil.safi2string(safi) + " " + tabRtrmapN.rd2string(ntry.rouDst) + " " + ntry.prefix);
        }
        if (!afiMsk(peerAfis, safi)) {
            return;
        }
        if ((safi == parent.afiUni) && (parent.flaps != null)) {
            parent.prefixFlapped(ntry.prefix, "null");
        }
        tabRoute<addrIP> learned = getLearned(safi);
        if (learned == null) {
            return;
        }
        if (learned.del(ntry)) {
            return;
        }
        currChg++;
        tabRoute<addrIP> changed = parent.getChanged(safi);
        if (changed == null) {
            parent.needFull = true;
            return;
        }
        changed.add(3, ntry.copyBytes(), false, false);
    }

    /**
     * process reachable prefix
     *
     * @param safi address family
     * @param ntry route entry
     */
    protected void prefixReach(int safi, tabRouteEntry<addrIP> ntry) {
        if (debugger.rtrBgpTraf) {
            logger.debug("reachable " + rtrBgpUtil.safi2string(safi) + " " + tabRtrmapN.rd2string(ntry.rouDst) + " " + ntry.prefix);
        }
        if (!afiMsk(peerAfis, safi)) {
            return;
        }
        tabRoute<addrIP> trg = null;
        if (safi == parent.afiUni) {
            trg = currUni;
        }
        if (safi == parent.afiLab) {
            trg = currUni;
        }
        if (safi == parent.afiMlt) {
            trg = currMlt;
        }
        if (safi == parent.afiFlw) {
            trg = currFlw;
        }
        if (safi == parent.afiVpnU) {
            trg = currVpnU;
        }
        if (safi == parent.afiVpnM) {
            trg = currVpnM;
        }
        if (safi == parent.afiVpnF) {
            trg = currVpnF;
        }
        if (safi == parent.afiVpls) {
            trg = currVpls;
        }
        if (safi == parent.afiEvpn) {
            trg = currEvpn;
        }
        if (safi == parent.afiMdt) {
            trg = currMdt;
        }
        if (trg == null) {
            return;
        }
        trg.add(3, ntry, true, true);
    }

    private boolean prefixReachable(tabRoute<addrIP> learned, tabRouteEntry<addrIP> ntry) {
        if (debugger.rtrBgpTraf) {
            logger.debug("processing " + tabRtrmapN.rd2string(ntry.rouDst) + " " + ntry.prefix);
        }
//////////////////////////
        if (neigh.enforceFirst) {
            switch (neigh.peerType) {
                case rtrBgpUtil.peerExtrn:
                case rtrBgpUtil.peerServr:
                    if (rtrBgpUtil.firstIntList(ntry.pathSeq, neigh.remoteAs)) {
                        return true;
                    }
                    break;
                case rtrBgpUtil.peerCnfed:
                    if (rtrBgpUtil.firstIntList(ntry.confSeq, neigh.remoteAs)) {
                        return true;
                    }
                    break;
                default:
                    break;
            }
        }
        if (!neigh.allowAsIn) {
            switch (neigh.peerType) {
                case rtrBgpUtil.peerExtrn:
                case rtrBgpUtil.peerServr:
                    if (rtrBgpUtil.findIntList(ntry.pathSeq, neigh.localAs) >= 0) {
                        return true;
                    }
                    if (rtrBgpUtil.findIntList(ntry.pathSet, neigh.localAs) >= 0) {
                        return true;
                    }
                    break;
                case rtrBgpUtil.peerCnfed:
                    if (rtrBgpUtil.findIntList(ntry.confSeq, neigh.localAs) >= 0) {
                        return true;
                    }
                    if (rtrBgpUtil.findIntList(ntry.confSet, neigh.localAs) >= 0) {
                        return true;
                    }
                    break;
                default:
                    break;
            }
        }
        switch (neigh.peerType) {
            case rtrBgpUtil.peerIntrn:
                addrIP a = new addrIP();
                a.fromIPv4addr(parent.routerID);
                if (ntry.originator != null) {
                    if (a.compare(ntry.originator, a) == 0) {
                        return true;
                    }
                }
                if (ntry.clustList != null) {
                    if (rtrBgpUtil.findAddrList(ntry.clustList, a) >= 0) {
                        return true;
                    }
                }
                break;
            case rtrBgpUtil.peerRflct:
                if (ntry.originator == null) {
                    a = new addrIP();
                    a.fromIPv4addr(peerRouterID);
                    ntry.originator = a;
                }
                a = new addrIP();
                a.fromIPv4addr(parent.routerID);
                if (a.compare(ntry.originator, a) == 0) {
                    return true;
                }
                if (ntry.clustList == null) {
                    ntry.clustList = new ArrayList<addrIP>();
                }
                if (rtrBgpUtil.findAddrList(ntry.clustList, a) >= 0) {
                    return true;
                }
                ntry.clustList.add(a);
                break;
        }
        if (neigh.nxtHopPeer) {
            ntry.nextHop = neigh.peerAddr.copyBytes();
        }
        if (neigh.dmzLinkBw >= 0) {
            if (ntry.extComm == null) {
                ntry.extComm = new ArrayList<Long>();
            }
            ntry.extComm.add(tabRtrmapN.dmzbw2comm(neigh.localAs, neigh.dmzLinkBw));
        }
        if (neigh.removePrivAsIn) {
            rtrBgpUtil.removePrivateAs(ntry.pathSeq);
            rtrBgpUtil.removePrivateAs(ntry.pathSet);
        }
        if (neigh.overridePeerIn) {
            rtrBgpUtil.replaceIntList(ntry.pathSeq, neigh.remoteAs, neigh.localAs);
            rtrBgpUtil.replaceIntList(ntry.pathSet, neigh.remoteAs, neigh.localAs);
        }
        learned.add(3, ntry, false, true);
        return false;
    }

}
